
// number of possible distributions
//  of p balls (particles) into n bins (de: Kaesten) [of depth d[]]

#include <stdlib.h>
#include <stdio.h>

#define MAXV 16  // max volume of bins or max 2s=16 spin=8

long int p_of_n(const int n, int p) {
 // max_int64: 20!=2.4e18 0x21c3h12
 //   2s=1,n=61 0x033ah12=232e15 2s=2,n=84 2s=3,n=99
 long int x=1; int j;
 if (p > n || n < 0) return 0;
 if (2*p > n) p=n-p; //  loop speedup
 for(j=1;j<=p;j++) x = x * (n+1-j) / j;
 // echo "DBG: p=$p n=$n p_of_n=$x" >&2
 return x;
}

// mixed bin depths = max digit general counter?
int maxp[/*depth=*/MAXV+1]; /* number for each particles-per-site config */ 
// 8*(2s=1): maxp[]={8,min(8,p),0,0,0}; // [0] unused count number free bins
// 8*(2s=2): maxp[]={8,8,8,0,0}; // use min or continue-loop 
// 8*(2s=3): maxp[]={8,8,8,8,0}; 
// 4*(2s=1)+4*(2s=2):
//           maxp[]={8,8,4,0,0}; 
// adapted_maxp: correct maxp[i-j] if cnt[i]+1 (speedup only?)
// 8*(2s=1): maxp[]={0,min(8,p),0,0,0};
// 8*(2s=2): maxp[]={0,min(8,p-2*[2]),min(8,p/2),0,0}; 
// 8*(2s=3): maxp[]={0,8-3[3]-2[2],min(8,p-3*[3]),min(8,p/3),0}; 
// 4*(2s=1)+4*(2s=2):
//           maxp[]={0,min(4+4,p-2[2]),min(4,p/2),0,0}; 

#define min(a,b) (((a)<(b))?(a):(b))

int main(int argn, char *argv[]){
  int v, n, p, j, // volume_bin, num_bins, num_particles, loop-vars
     i[MAXV+1]; // loop indizes for-loops
  int is; // sum bins filled partly   
  long int x, y[MAXV+1]; // sum_result, binomial coeffs subspaces
  long int py; // used for product y[i]
  printf("computing num distributions of\n"
    " p balls into n bins of size v*balls (for v=1: binomial coeff.)\n");
  if (argn != 4) { // arg0 + 3 args
    printf(" options: <v=1,2,3,...,16> <n=1...128> <p=0...v*n>\n");
    return 1;
  }
  v = atoi(argv[1]);
  n = atoi(argv[2]);
  p = atoi(argv[3]);
  // printf("L%d\n",__LINE__);
  printf(" v=%d n=%d p=%d\n", v, n, p);
  if (v<1 || v>MAXV || n<1 || p<0 || p>v*n) {
     printf("ERR: bad args (x=0)\n");return 2;};
  x=0;
  if ( 2*p > v*n ) p = v*n - p; //  faster loop using (n p) == (n n-p)
  // !!! keep it simple !!! the outer loop
  // for p=2s=2 we have binomial coeff., Pascal's triangle
  // for p=2s=3 we have product of 2 binomial coeff.
  // index i[v] means number of bins filled with v balls
  // break v-dim-loop if highest loop counter is reached i[v..2]=max,...
  // ini-v-dim-loop fill low-significant-subcounter i[0] first
  is=p; for (j=1;j<=v;j++) { i[j]=min(n,(is+j-1)/j); is -= i[j]*j; }
  fprintf(stderr, "DBG: ini-fill-type-counter: ..");
  for (j=1;j<=v;j++) printf("%d%c",  i[j],((j==v)?'\n':'-'));
  if (v==3) // TODO all v's
  for (i[3]=0;i[3]<=p/3;i[3]++) // start most significant idx: 3p-bins
  for (i[2]=0;i[2]<=(p-3*i[3])/2;i[2]++) // num 2-particles bins (sites)
  { // v-dim-loop
    // i[1] = p-3*i[3]-2*i[2];  // num bins with 1 particle per v=3-bin
    // i[0] = n-i[1]-i[2]-i[3]; // num empty bins, may become negative
    i[1]=p; for (j=2;j<=v;j++) i[1] -= j*i[j]; // compute left 1p-bins
    i[0]=n; for (j=1;j<=v;j++) i[0] -= i[j];   // compute left 0p-bins
    // n0f=$i3; n1f=$i2; n2f=$[p-3*i3-2*i2]; n3f=$[n-n0f-n1f-n2f]
 //    if ( i[0] < 0 ) continue; // i[0] may become negative by above loop-fn
//  y[3]=p_of_n( n, i[3]);           // num possible triple-spin-sites
//  y[2]=p_of_n( n-i[3], i[2]);      // num possible double-spin-sites
//  y[1]=p_of_n( n-i[3]-i[2], i[1]); // num possible free-sites in rest
    is=i[0];
    for (j=1;j<=v;j++) { is += i[j];
      y[j]=p_of_n( is, i[j]); } // binomial coeff, num variants filled subset
    fprintf(stderr, "DBG: fill-type,coeff: ");
    for (j=0;j<=v;j++) printf("%d%c",  i[j],((j==v)?' ':'-'));
    for (j=v;j>=1;j--) printf("%ld%c", y[j],((j==1)?'\n':'*'));
    py=1; // may overflow
    for (j=1;j<=v;j++) py = py * y[j]; // product y[1]*y[2]*...*y[v] 
    x = x + py;
    // break v-dim-loop if highest loop counter is reached i[v..2]=max,... 
    is=p; for (j=v;j>=2;j--) { is-=i[j]*j; if (is >= j) break; }
    if (j==1) printf("DBG: max-v-dim-loop-counter reached, stop\n");
  } // v-dim-loop
  printf("Result: x= %19ld 0x%016lx\n", x, x);  
  return 0;
}
