#ifndef hilbert_h 	/* include this only once */
#define hilbert_h 1

/* definition of the base and the operators of the Hilbert space */

#include "modeldef.h"
#include "defs.h"
//#include "h_file.h" /* do not include upper-level headerfiles */
#include "symmetry.h" /* t_sites */
//#include "spins.h"  // do not include upper-level headerfiles

/* t_sites and tphase moved to symmetry.h */
/* tbase moved from spins.h to hilbert.h 2017-01-23 */
/* smallest type for base configurations dN..d1 uN..u1 */
/* cray: sizeof() short=int=long=longlong! but bits are different!? */
/* 1998-148: cray (short)0100=0000 (short)0080=ff80 !?
   1998-150: tbase1=1bit/site, tbase = CM ( tbase1, tbase[u,d], .. )
   2002apr: Nd,...,3d,2d,1d,Nu...,3u,2u,1u in one word
   ToDo: move to hilbert.h (tbase.h is only for l1[])
   BitsPerSite=CM(1,2,2)
   ToDo: rename tbase (more physics name) to tbits (more computer science)
       think, that "tbits" will be better undertstud by new developpers
 */
#ifndef __STDC_LIMIT_MACROS /* defined in gcc6, but not in gcc4 2017-01-25 */
#define __STDC_LIMIT_MACROS 1 /* ISO C99 C++ wants it for UINT64_MAX */
#endif
#include <stdint.h> /* standart integer, uint64_t on gcc-m32 2015-05-11 */

/* ToDo17: define a string for debug output how tbase is defined */
/* ToDo17: set an macro if tbase is int or array|class ++ possible? */
#if   CM(1,2,2)*NN <= 8*S_int         /* 32bit */
  typedef unsigned int       tbase1;
# define S_tbase S_int
# define dbg_tbase "uint"     /* for debugging output of tbase type */
#elif CM(1,2,2)*NN <= 8*S_long        /* 32 or 64bit CPU+cc-dependent */
  typedef unsigned long      tbase1;
# define S_tbase S_long
# define dbg_tbase "ulong"     /* for debugging output of tbase type */
#elif UINT64_MAX        &&  CM(1,2,2)*NN <= 64
/* 
  tested HP-mini Atom tU20 4 2 -2 4 6+20 n1=464 p1=1 p6=1 -11.70820390 2015-05-17
  long64: minsymcfg_lNbrk t=1.1s loops/s= 4.7e+05 t[ns]/cfs=   26.68 0 ud=6+20
  long64: minsymcfg_loopN t=1.1s loops/s= 5.9e+04 t[ns]/cfs=  211.93 0 4Ni
  uint64 == long64
  versus old vlint: (10..13 times slower, 2019.03 bad bfly implem.? ToDo)
   vlint: minsymcfg_lNbrk t=1.4s loops/s= 4.8e+04 t[ns]/cfs=  261.67 0 ud=6+20
   vlint: minsymcfg_loopN t=1.8s loops/s= 4.5e+03 t[ns]/cfs= 2796.99 0 4Ni 
   after some vlint-optimizations factor 8 (ToDo: further vlint-speedup possible)
*/
   /* test 2015-05-11 gcc412-m32 NN=40 */ 
  typedef uint64_t           tbase1;
# define S_tbase 8   /* in Bytes! */
# define dbg_tbase "uint64"     /* for debugging output of tbase type */
//#elif  defined(_GLIBCXX_USE_INT128) &&  CM(1,2,2)*NN <= 128
#elif 1 && defined(__SIZEOF_INT128__) &&  CM(1,2,2)*NN <= 128 
/*  gcc4.7 -m64 ok (-m32 not, gcc44 not)
  gcc -mtune=native -dM -E -x c /dev/null | grep INT128
     # __SIZEOF_INT128__ 16
  configure2015: test __int128 and set -D__SIZEOF_INT128__=16 or -DNO__int128
        ToDo18: ... if not already set by cc/CC itself
  bits/c++config.h
    define:  _GLIBCXX_USE_INT128
             _GLIBCXX_USE_FLOAT128
    defined: _GLIBCXX_LONG_DOUBLE_COMPAT __LONG_DOUBLE_128__
    included by: complex.h cxxabi.h fenv.h tgmath.h
  tested on HP-mini Atom tU40 2015-05-17 2+38=3894 3+37=6e5(0m2) 4+36=5e7(5m++)
 __m128i: ... compile errors ... ToDo howto use 128bit-mmx-x86-registers?
    failed, getl1-alignment?
 __int128: minsymcfg_lNbrk t=1.7s loops/s= 1.5e+05 t[ns]/cfs=   41.66 0 ud=6+34
 __int128: minsymcfg_loopN t=1.1s loops/s= 7.2e+03 t[ns]/cfs=  869.83 25 4Ni
 versus old vlint: (24..90 times slower)
   vlint: minsymcfg_lNbrk t=1.4s loops/s= 2.3e+04 t[ns]/cfs=  269.36 34 ud=6+34
   vlint: minsymcfg_loopN t=2.2s loops/s= 9.2e+02 t[ns]/cfs= 6768.18 44 4Ni
*/
//#include <emmintrin.h>  /* __SSE2__ */
// ToDo18: set Tnorm2=__int128 too? test = ok
//  typedef __m128i            tbase1;
// ToDo: configure only? not on gcc -m32
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wpedantic" /* 2018-11-11 disable i128 warns */
  typedef unsigned __int128    tbase1; /* ToDo: check by configure */
#pragma GCC diagnostic pop

# define S_tbase 16  /* Bytes! 2016-01 was 128 */
# define dbg_tbase "uint128"     /* for debugging output of tbase type */
#elif 0 & defined(__SSE2__) &&  CM(1,2,2)*NN <= 128 
/*  gcc4.7 -m64 ok (-m32 not, gcc44 not)
 __m128i: ... compile errors ... ToDo howto use 128bit-mmx-x86-registers?
 src/tbase.h:39: error: incompatible types when assigning ... from type long
  int a[] __attribute__ ((aligned (16))) = {0,0,0,1};
  no shift operations on SSE2
 __m128i t = *(__m128*)a, x= __mm_XXX(t); ???  */
#include <emmintrin.h>  /* __SSE2__ */
/* direct access to 128bit registers 
  __m128 for float, __m128d for double and __m128i for int, short, char */
  typedef __m128i            tbase1;
# define S_tbase 16 /* Bytes!, 2016-01 was 128 */
# define dbg_tbase "sse2.m128"     /* for debugging output of tbase type */
# define SSE2_tbase 1  /* every tbase access needs its own __mm-functions? */
#else 
# if 0  /* old non log2 but packed, not benes compatible 2015-12 */
/* ToDo: use __uint128_t or 256bit vectorregisters if <<>>|& is defined */
/* 2015-09-23 non-log2-types do not work for benes! */
#  define B2size ( CM(1,2,2)*((NN-1)/(8*S_int)+1)) /* size in int =fastest */
#  define S_tbase (B2size * S_int)  /* usually Bsize <= (B2size*S_int) */
# endif
# if   CM(1,2,2)*NN <= 128 /* added 2015-12-31, benes needs 2^n */
#  define S_tbase 16  /* Bytes!, 2016-01 was 128 */
# elif CM(1,2,2)*NN <= 256 /* added 2015-12-31, benes needs 2^n */
#  define S_tbase 32  /* Bytes!, 2016-01 was 256 */
# elif CM(1,2,2)*NN <= 512 /* added 2015-12-31, benes needs 2^n */
#  define S_tbase 64  /* Bytes!, 2016-01 was 128 */
# elif CM(1,2,2)*NN <= 1024 /* added 2015-12-31, benes needs 2^n */
#  define S_tbase 128 /* Bytes!, 2016-01 was 128 */ 
# else
#  error "needs new definitions for more than 1024 bits"
# endif
# define B2size (S_tbase / S_int)
# define dbg_tbase "cxx.vlint32"   /* for debugging output of tbase type */
# if (B2size * S_int) != S_tbase  
#  error "S_tbase is not a multiple of S_int, fatal"
# endif
#ifdef __cplusplus
#include "vlint.h"     /* speed loss! avoid this! sets _Vlint_h = 1 */
#else
#  if   CM(1,2,2)*NN <= 128
#    error "need c++.vlint.h for bigger tbase OR gcc -D__SIZEOF_INT128__=16"
#  else
#    error "need c++.vlint.h for tbase bigger than 128 bit"
#  endif
/* ToDo: replace by functions, setzero, set/getbit, shift, ...? */
/* removed v2.38 for better MPI-compatibility, problem for gcc -m32 NN>32 */
/* use structure of bitfields ??? + setsite, getsite */
/*#  include "/dev/zero" */ /* boese! */
   typedef int_break_cc bad_syntax[##];
#endif  /* if __cplusplus */
typedef class vlint tbase1;   /* unsigned int[] */
#endif  /* CM(1,2,2)*NN does not fit into any int type, to long */
typedef tbase1 tbase;

/* definition of the vakuum state */
#define TBASE0 ((tbase)0)

/* ERR(600): try one of unsigned char, short, int, long, double(?) */
/*            ===========> consumes more memory  */
#ifndef ATnorm2 /* ToDo18: check if needed, size of order nn^2 ? */
#define ATnorm2 unsigned char	/* range=0..255^sizeof(Tnorm2), for arrays */
#endif
#ifndef Tnorm2        /* should be same or bigger than Tbinom */
#define Tnorm2 ULong  /* prefactor, default=ULong (try __int128 or double) */
#endif
#ifndef Tbinom /* type binomial coefficient  ULong or bigger see np() */
/* max_nCk(nn,nu): i32=(34,17) i64=(67,33) i128=(131,65) f64=(1020,510) */
/* int64 is faster than int128, float64 has lower initial accuracy (2018) */
/* using uint32 or smaller-than-int types makes only sence for debugging */  
/* Tbinom must be unsigned (ULong = unsigned long int, unsigned __int128) */
/*  can be defined outside by -DTbinom=unsigned\ __int128 */
/* examples: N=3 2s=7 nn=21 warn+err9110 at ui16+ui16 */
#define Tbinom ULong  /* ULong=__int64, __int128 or double, see binomial_ini */
#endif

/* ------- class base_ => b_ ----------------------------------- */
/* tJ and tU still needs get_symbase and MaxSBase>1 (this is slow on MPI) */
/* symmetric base structure (single element, no block, slow)
 * v1: buffer and results for minsymcfg function,
 * ToDo: [MaxSBase abtrennen] AH zufuegen? sbase + fullsbase[1]=Slow
 */
struct sbase{	/* memory intensive, how archive locality? [AH]! */
  tbase  scfg;	/* minsymcfg, old: b_smallest */
  Tnorm2 norm2;	/* norm^(-2),   <cfg|scfg),  N=10 S=5/2 norm2>1e9 */
  tphase sign;	/* phase, real=int=+1,-1 */
  /* ToDo: remove */
//  int    num;		/* number of configs (mostly 2*MaxSym+2) (needed?)*/
  /* ToDo: remove */
//  int    o;		/* option: 0=fullbase,1=minsymcfg,2=isminsymcfg */
  /* waste nhm space ToDo: use malloc or separate split off full_sbase */
  /* used in b_out() b_twice allsymcfg_lm allsymcfg_ud allsymcfg isminsymcfg_slow
   *         ini_norm2(DBG) xval.c.modeloverlap() stringorder[2] 
   */
};

//#if defined(CONFIG_MPI) && CM(1,0,0)
//#define MaxSBase  1  /* set to mimimum acceptable size, less debugging */
//#else
#define MaxSBase (2*MaxSym+2)	/* should be larger than 2*MaxSym+1 */
//#endif
/* full symetric base structure (2015-03-10) old: sbase
 * v1: buffer and results for minsymcfg function,
 * ToDo: [MaxSBase abtrennen] AH zufuegen? sbase + 
 *  fullsbase[1]=Slow for (slow) debugging and test
 */
struct fullsbase{	/* memory intensive, how archive locality? [AH]! */
  tbase  scfg;		/* minsymcfg, old: b_smallest */
  Tnorm2 norm2;		/* norm^(-2),   <cfg|scfg),  N=10 S=5/2 norm2>1e9 */
  tphase sign;		/* phase, real=int=+1,-1 */
  int    num;		/* number of configs (mostly 2*MaxSym+2) */
  int    o;		/* option: 0=fullbase,1=minsymcfg,2=isminsymcfg */
  /* waste nhm space ToDo: use malloc or separate split off full_sbase */
  /* used in b_out() b_twice allsymcfg_lm allsymcfg_ud allsymcfg 
   *         ini_norm2(DBG) xval.c.modeloverlap() stringorder[2] 
   * ToDo: alloc ib,is locally! to spare mem
   */  // TODO remove MaxSBase-need (memory waste)
  tbase  ib[MaxSBase];	/* ib[MaxSBase] large object -> class for array proof? */
  tphase is[MaxSBase];  /* NULL if empty/unused, malloc(MaxSBase*..) needed */
};

/* AH NZX VS shbuf from spins.h to h_file.h(matrix-store-defs) 2017-01-23 */
/* ToDo: alloc in dynamic way and store max needed size!?
 *       better fix for MPI_transfers (but fill with mpi*n Hlines[blk]
 * NZXMAX max# of elements of one line in a HBlock, smaller is faster
 *   see output of h_nz_x_max for a optimum value
 * NUM_AHEAD_LINES is an buffer to store H lines ahead for less MPI calls
 * 2017-01-13 64 is to bad scaling see y2016/lrzGarching*.txt AH=32K HBLen=4M
 *          1024 used now (eats more memory, p.e. 8M growed to 41MB)
 *          see plot speed/core vs. cores (using min. nodes for growing n1)
 * 2018-11 if NAH*B_NUM to big, and bigger XY_TYPE, linker errors, old 1024
 *     check: nm --size-sort -S ./spin         # for NN=256 AH=1K BNUM=2
 *            nm --size-sort -S ../src/spins.o # 36009980 B nrecv_cfg 1GB FAIL
 *     Todo18: dynamic allocation of nrecv_vy etc.
 */
#ifndef  NUM_AHEAD_LINES  /* better cache usage, but bigger buffer size */
# define NUM_AHEAD_LINES (1024)   /* must be 1 or greater, static memory! */
#endif

/* VS is for unroll/vectorization/pipeline-filling = small number 2^x
   optimum depends on SIMD-vector size, pipelinedepth, num ALUs etc.
   see hilbert.h table for best choice
   2016-01-09 best VS depends from NN and CM() !!! 4*64bit 20, ToDo: check!
    gcc -march=native -x c -E -dM ../src/hilbert.h | grep AVX # AVX+AVX2
    S_tbase = S_long = (__WORDSIZE/8) = (64/8)
    tuning for vector size VS (depends on CPU ALUs for integer, shift etc.)
*/
#ifndef VS                /* best use cc-option to change VS: -DVS=4 */ 
#if   defined(__AVX2__)   /* best found for HaswellEP-E5-2630v3 is VS=32 */
                /* but EPYC best is VS=8, ToDo19: needs improvements */
#define VS (32*8/S_tbase)
#elif  defined(__tune_atom__) /* best for N455 is VS=8or16 but HT.VS=8 */
#define VS (8*8/S_tbase)
#elif  defined(__SICORTEX__) || defined(_MIPS_ARCH) /* best is VS=4 */             
#define VS (4*8/S_tbase)
#else
#define VS (8*8/S_tbase)   /* best for SSE2? */
#endif /* AVX2? */
#endif /* undefined VS */

#if VS > NUM_AHEAD_LINES && VS > 4 && NUM_AHEAD_LINES >= 4
#undef  VS
#define VS 4 /* 2018-11-11 for small AH testing, AH must be multiple of VS */
#endif
#if VS > NUM_AHEAD_LINES && VS > 2 && NUM_AHEAD_LINES >= 2
#undef  VS
#define VS 2 /* 2018-11-11 for small AH testing, AH must be multiple of VS */
#endif

#if (VS <= 0)
#error bad VS
#elif (VS > NUM_AHEAD_LINES)
#error VS bigger NUM_AHEAD_LINES
#endif

#if (VS > NUM_AHEAD_LINES) || (VS <= 0)
#undef  VS
#define VS 1 /* 2018-11-11 for small AH testing, AH must be multiple of VS */
#endif

/* --------    packed-struct-array (2019-01 moved from h_file.h) ------- */
/*
  if HRMAX greater than 0 (compressed mode),
  the program stores a index to the double or complex value of the matrix
  element, else the matrix element itself is stored to the disk
 --- if an error apears, choose xy as a next larger type ---
 --- please edit HRMAX too !!! ---
 HRTYPE: unsigned char , unsigned int , ULong
 * HRMAX int-index: 1B..4B may overflow, less Matrix mem,
 *   faster IO, if nzx.rr transferred, ToDo19: transfer v1[rnd_j] only
 * HRMAX float: 4B..16B, reduced accuracy for less mem
 *   but more robust + if ToDo19 above same speed, simpler code
 * speed_mpi: ToDo
 *  ToDo19: how to compute HRMAX early and fast?
 *     rnd_cfg 4*numJ*numSymOrbits^2 big on highS
 *  ToDo: examples: J1sq40d14tsk32.hrmax= 23  diagZiZj+2nondiag?
 *  J1sq40 160sym 38+2.hr=                                  +10=22-12 -t2
 *  J1sq40 160sym 38+2.hr=7-2  +4=14-5   +6=15-8   +8=20-11 +10=26-13 -t1
 *                         5        9        7          9       13
 *  J2sq40   1sym 38+2.hr=5-3  +4=17-15  +6=36-34  +8=57-55?
 *  J2sq40   2sym 38+2.hr=9-3  +4=21-15  +6=40-34  +8=61-55?
 *  J2sq40  40sym 38+2.hr=13-3 +4=29-15  +6=44-34  +8=69-55
 *  J2sq40 160sym 38+2.hr=13-4 +4=33-15  +6=48-34  +8=73-55 +10=83-71?
 *                          9      18        14        18       12
 *  Jn=lot-diag-IsingErgs, few-nondiag
 *  2sJ2sq40:   1sym   +2=5-3  +4=19-17  +6=36-34
 *  2sJ2sq40: 160sym   +2=13-4 +4=35-17  +6=58-44
 *                          9      18        14   
 *          counts diag only (nzx=0(+diag) as nondiag?)
 *  max. (N-1)*N J's = max N^2 diff. NZX terms per cfg, 2^(Nw) ising-ergs?
 *    pot. max N^2=Nw non-diag-ergs (ggf. zu N*N erweiterte isingterms)
 *             *symorbits^2 = Nw*...   ... HRMAX=0 safer
 * ToDo: divide into diag.hr (store direct) + nondiag.hr (small? faster srch?)
 */
#ifndef HRMAX		/* can be defined in config.h */
/* ToDo18: HRMAX=0 is the most robust form, but it eats 50-300% more memory */
#define HRMAX   0  /* 255, 1023(old) or higher, if ERR(9100), v2.33 */
#endif

#ifndef HR_TYPE  /* depends from HRMAX, edit this too */
/* ToDo2015: IEEE 754-2008 16bit-Minifloats ...
 *    16bit half precision 1sig+5exp+10mant(+1)  lg(2^11)=3.31 dec digits
 *     alternativ: bfloat = 1sig+8exp+7mant(+1)   lg(2^8)=2.41 dec digits
 *    float: 1sig+8exp+23mantissa(+1)            lg(2^24)=7.22 dec digits
 * ToDo2015: or 3bitparam*3bitn*0.5*sqrt(5bitn1/5bitn2)=16bit
 * ToDo18: HR_TYPE = scplx if HRMAX > 256^sizeof(scplx)
 *  see spins.c hamilton_puth_line.HRMAX + ...
 */
#if (!defined(HRMAX)) || (HRMAX<=0) 
#define HR_TYPE scplx           /* short cplx (minsize), see defs.h */
#elif   HRMAX<256
#define HR_TYPE	unsigned char	/*  8bit char  sz 256 */
#elif HRMAX<65536
#define HR_TYPE	unsigned short	/* 16bit short sz 65536 */
#else   // if 65536<=HRMAX<2^32
#define HR_TYPE	unsigned int	/* int32 sz 256^4, reel float id better */
#endif  // if HRMAX<256
#endif  // ifndef HR_TYPE
// ToDo18: index table generation before H-start? using seperate norm2-tab?
//         and per NZ-param?

/* y-index + next-x-flag of H element */
#ifndef BY_TYPE                 /* allows -DBY_TYPE="unsigned int" outside */
/* 2018-05 ushort triggered by 2048nodes*2tasks/node*16threads=65536threads */
/*    but since 2.36 we have to put max_mpi_i only, not num_threads */
/*    means it must be triggered for mpi only using 65536++ tasks */
/*  "int" would allow 2e9 threads or "unsigned int" 4e9 threads 2018-05 */
#define BY_TYPE unsigned short  /* char,short,int = lg2(blocks*mpi) v2.33 */
// ToDo18: BY for MPI*PT or MPI only? concflicts with XY-description!
//         test with uchar, max 255 ? abj=0..(pt_n*mpi_n-1)?
//      nonMPI does not need by or BY_TYPE? mpi_only=16*16>255
//       bad warning and error? v2.33 was mpi*pt-blocks, since 2.36 mpi-blocks
#endif /* BY_TYPE */
/* add DFLAGS+= -DXY_TYPE=ULong in Make_cfg.inc if uint is not enough */
#ifndef XY_TYPE                 /* allows -DXY_TYPE=__int128 outside */
#define XY_TYPE unsigned int	/* Ulong, needes more memory + IO! */
         /* 32bit is enough for 2G-elements to 8B = 16GB/MPItask */
         /* if XY_TYPE is triggered you can also increase MPI tasks */
         /* for 32GB/node no overflows with minimum 2 tasks/node */
#endif /* XY_TYPE */
/* XY_FLAG: bit in XY_TYPE = last element in H-lines, ToDo: better num-list */ 
#define XY_FLAG (((XY_TYPE)1)<<(sizeof(XY_TYPE)*8-1))


//#ifndef AddSS
//#define AddSS 0 /* 0=default or float-number: H+S^2 */
//#endif

/* NZXMAX = maximum number of non-zero elements per matrix line
 *   keep it as minimum to minimize buffer size (NUM_AHEAD_LINES*NZXMAX) 
 * 2015-05-11 NZX=multiple_of_32=(..+31)&~31 for VS-vectors at AH=1
 *            but its vasting mem, better use AH multiple of 32
 * 2017-04 multiple 32 removed, tJ + tU case added, ToDo: lower upper bond
 */
/* this is for lattices with fixed number of neighbours per sites only:
 * old2016: max. 1 nondiag-term per bond (tU: 2/bond)
 * new2017: min(particles,holes)*max_siteneighbourbonds
 * JJ: maxdim for nu=NN/2
 *     terms=min(nu,nd)*(Nw/NN)+1*diag
 *     u-d or d-u = 1 non-diag-term
 * tJ: u-d or d-u or .-d or .-u ... = 1 non-diag-term (t or J, never both)
 * tU: maxdim for nu=nd=NN/2
 *     terms=(min(nu,NN-nu)+min(nd,NN-nd))*(Nw/NN)+1*diag
 *     .-3 = 2 non-diag-t-terms (.3 to ud+du+3.) + 1 JJ-term
 *            (or seen as .-u + .-d with 1 t-Term each + both hops together)
 *            + both-hopping-term (JJ or S+S-) = max 3 terms per bond
 * 2017-03 old upper bound NZXMAX=Nw+1, is not the minimum upper bound
 *      j1-j2-square Nw=(4*NN)  NZXMAX=4*NN+2   NN=40: 192
 *                           newNZXMAX=4*NN/2+1 NN=40:  81 (43% Memory)
 * ToDo17: make a Nw-check at start (big Nw causes big NZX + vasting memory)
 * ToDo17: make a NZXMAX-check at start + report possible to low NZXMAX
 */
/*  AddSS: what is the matrix non-zero density for this case?
 *     JJ NN=40  (20^2=400) / ((40 over 20)=138e9) is very sparse matrix
 *  JJ: all-to-all sites max(nd*nu) + JJ-bond-terms
 *  tJ: all-to-all sites max(nd*nu) + (t-bond-term xor JJ)
 *  tU: all-to-all sites min(nd,dholes)*min(nu,uholes) + 2t-bond-terms + JJ
 */
#ifndef NZXMAX
# ifdef AddSS    /* float constant for all-to-all operator (sumSi)^2 */
#   define NZXMAX (((NN)/2)*((NN)/2) + (Nw)*CM(1,1,3)+2)
# else
#   define NZXMAX                    ( (Nw)*CM(1,1,3)+2)
# endif
#endif

/* number of elements per hnz-read-buffer (HBLen*(XYsize+HRsize))
 * 2017-01-23 MPI-I/O is the bottleneck, dont care about CPU caches here
 *  HBLen = NUM_AHEAD_LINES * NZX  to have same MPI-packetsize in SH + i100
 *  ToDo:  but to big using NZXMAX? do malloc, remove HBLen, AH=2**n
 */
//  16.02.17 HBLen now depends from AH*NZX, synchron nonequal SH blocks
//           old: asynchron equal SH blocks, but buggy complex MPI sync
//  HBLen = H-matrix block length == nzx*NAH
#define HBLen ((NZXMAX)*(NUM_AHEAD_LINES)) /* 2017-01-23 optimized for mpi */ 
/* structure to hold stored H-matrix elements
 * direct access to nonaligned members cause trap on IRIX ([].xy, [].r)
 *   therefore xy[],r[]
 * Compressed sparse row (CSR, CRS or Yale format) ofs by XY_FLAG
 * ToDo17 store data sorted to destination nodes!
 * ToDo18 move into thbuf=struct shbuf (fat) to eliminate global b_hbuf
 */
struct thxy {	/* y,h[x,y] if(x=y)x++; compact sparse matrix storage */
   /* ToDo: add xx (Hsize+30%) and sort to mpi-blocks (remove by),
            simpler mpi code (noXY_FLAG),
            faster mpi (const HBLen*XY_TYPE=mpi_pckt_size)?, +traffic/2
            xx,rr stored local, xy stored at remote node */
   /* extra data can be stored on first diag element xx and xy */
   /* XY_FLAG=1bit*NZX*AH z.B. 33bit*n1, so extra ofs[AH] (2B) eats less */
   /* 2017-01-24 xy_ofs[] added to replace XY_FLAG + HBLen by NZX*AH */
   // 2017-02 better extra xy_ofs[node_n1], step 2 struct_shbuf[HBLen] 
#if 0  // for future use, replacing end of line bit, see nzofs below
#if HBLen < 65536 /* 2B=16bit-ushort or 4B=32bit-uint ofs[0..l-1]+len[l] */
   unsigned short xy_ofs[NUM_AHEAD_LINES+1];
#else  // HBLen do not fit to short=32bit 
   unsigned int   xy_ofs[NUM_AHEAD_LINES+1];
#endif // HBLen do not fit to short=32bit 
#endif // if 0 (for future use)
   /* ToDo.Problem {xy[],rr[]} has fixed max size, bad to store 
    * {xy,rr}[] may waste Bytes for alignment, but flex. size, by as xy-bits?
    *    ToDo: warn bad nzx, redesign later? (always some copying needed?)
    *   by could be recomputed log2(mpi_n)-rnd-read in mpi_n*8B 
    *    mpi_n=100k 17*rnd-read on 800kB
    *   xy+by as (by << logmaxxy) + xy 
    *  ToDo17: [HBLen].{[by,]xy,rr} to simply cut stored elements at nzx
    * define switch-Macro:
    *    #define M1(aaa,rr,idx) aaa.nz[idx].rr
    *    b_hbuf[blk].rr[ b_bnz[blk] ]=...     ... = tbh->xy[ ii ]
    *    b_hbuf[blk].nz[ b_bnz[blk] ].rr=...  ... = tbh->nz[ ii ].xy
    */
#if 0  /* old format, does not allow to easy store partly filled tables */
   /* ToDo19 delete, can not be used with now partly filled memHwrite */
   XY_TYPE xy[HBLen]; /* column y of element h[] (4B=32bit idx + XY_FLAG) */
   BY_TYPE by[HBLen]; /* block y  of element h[] (2B=16bit mpi node) */
   HR_TYPE rr[HBLen]; /* [index to] NZ-Element (uchar,2B=short,int,scplx) */
#else
  // TODO by shold be replaced by sorted xy,rr + num_nz_per_mpiblock[mpi_n]
  // TODO   no packed problem if stored separately
  //  g++4.1.2 "error: cannot bind packed field" ... "to 'const sComplex&'"
  //   but gcc4.9.3=OK HR16B  c128: 16+4+2=22B/nz (not_packed=24B)
  struct { /* 2017-02-20 does allow to store partly filled tables memHwrite */
    XY_TYPE xy __attribute__((__packed__)); /* column y of H (32bit idx+XY_FLAG) */
    BY_TYPE by __attribute__((__packed__)); /* block y  of H (16bit mpi node) */
    HR_TYPE rr  /* [index to] nzH (uchar,short,int,scplx) */
#  if defined(__GNUC_MINOR__) && __GNUC__ * 100 + __GNUC_MINOR__ >= 409
               __attribute__((__packed__)); /* gcc412-bug?[index to] nzH (uchar,short,int,scplx) */
#  endif
    ;
  } nz[HBLen];
#endif
   /* ToDo: better variant b or c ???
      variant a) all elements + flag EOL (or better xy_ofs)
      variant b) const number of nondiag elements (max hnzline)
      variant c) diag (has coded num nondiag) + nondiag
     used to write blocks to disk/storage but also should fit to CPU cache
    */
}; /* struct thxy = compact sparse matrix block-storage */

#define Hsize sizeof(struct thxy)

#if ((NZXMAX)*(NUM_AHEAD_LINES))%(VS) != 0
/* usually NUM_AHEAD_LINES is multiple of 1024, so this will never hit */
#  error "NZXMAX * NUM_AHEAD_LINES must be multiple of VS"
#  error "set NUM_AHEAD_LINES to multiple of VS, p.e. 1024*VS"
#  error "this allows auto-vectorization of minsymcfg_blk-code"
#  include "abort_with_error_here_if_not_done_by_error_above"
#endif
/* temporary buffer size for non-zero matrix elements of AH lines */
/* round up to multple of VS for vectorized hilbert.minsymconf... */
// #define HBLrVS (((NZXMAX*NUM_AHEAD_LINES+VS-1)/VS)*VS) /* 2015-06-01 */
#define HBLrVS ((NZXMAX)*(NUM_AHEAD_LINES)) /* 2017-01-24 simplifyed = HBLen */
/* --------------- this is (big dyn.) buffer for pthread code ------------- 
 * all arguments needed by sub-functions to calculate |H|m> or <n|H|v0> (?)
 *  within threads are stored here = thread local memory
 * 2017-01-28 make better readable longer names (grep++)
 */
typedef struct shbuf {	/* = thbuf, buffer to calculate a H-line H|icfg> */
  int   block; /* compute thread_idx or vector-blockNo (0..B_NUM-1) */
  ULong idxAHblk; /* 2017-02 index AH-block = 0..node_n1/AH */
  int   lenAHblk; /* nAH for src_range=i...i+nAH, mostly NUM_AHEAD_LINES */
  // ToDo why ULong? int is enough for AH
  ULong ofsAHblk; /* idx_of_in_cfg -b_ofs[] -mpi_ofs[], multiple of AH */
  // index_node_jj = idxAHblk * NUM_AHEAD_LINES + ofsAHblk; // 2019-05
  //   old version ofsAHblk was simply idxAHblk * NUM_AHEAD_LINES (?)
  //   new version ofsAHblk is used as operation index in nhm()
  scplx *v0, *v1; /* vector for <nHv> (v1[i..+AH]+=H*v0[nzi]) or NULL */
  struct sbase sci, scj;/* working sym-cfg space for nhv+nhm within threads */
     /* sbase = { scfg, norm2, sign }  */
     /* Todo: rename to scj = struct or sym cfg i+j ? old: b=base_cfg */
  /* mips multithreads reading same memory = slow! L1miss, ...
     ... local cache sym.ln[NN] sym.i[][NN] .w[] .numall 
     sym_ud? .wud nu nd */
  int nn; /* local copy of global nn, MIPS-multithread! 2017-09 */
  struct tsym *psym; /* pointer to global sym or local copy of sym 2017-09 */
#if defined(_MIPS_ARCH) && defined(CONFIG_PTHREAD) /* SiCortex? bad multithread */
  /* faster on multithread MIPS but cost more (read only) memory */
  /*   may be bad for shared caches (overflow by multple identical entries?) */
  struct tsym lsym; /* local copy against slow multithread read of same mem */
#endif
  tbase tcfg; /* target config for op_xx(cfg) in xval.c v2.36 obsolete? */
  double rx;  /* rx=<v0|H|v0> incl. nHn, geth_block, expect-value, obsolete? */
  void *iostruc; /* struct input paras + output results, see mrule() JS1904 */
  int line_start; /* used for puthbuf line compression for num_ahead>1 */
  double rtime; /* PRF: calculation runtime of this block, dbl 2013-04 */
  struct shbuf *hbuf_array; /* to optional access other thread rtime*/
  /* ToDo17 replace line_start by ofs_nz_i[AH+1], replace XY_FLAG */
  int num_nzAH; /* num nonzero-elements for AH matrix lines, 0<=nzAH<HBLrVS */
  int nzAH_overflow; /* on NZX*AH overflow count this up, debugging per block */
  int max_nzAH_overflow; /* on NZX*AH overflow give minimum needed nzAH */
  int max_nzAH_recv; /* estimate better BFACTOR by NZX <= 2*bounds*nu+1 ??? */
  // thbuf *hbuf_array; /* to optional access other thread rtime*/
  /* = diag+offdiag is more simple code, but +1/nzx wasted CPU */
  /* next two lines are for speed_test and simpler code (better+faster?) */
  /* +1 for simpler overflow-check cfg[lenAH]=zero triggers alarm-checks */
  tbase  cfg[NUM_AHEAD_LINES+1]; /* 2015-10-08 JS op_target_cfg_i[AHEAD] ??? */
  int  nzofs[NUM_AHEAD_LINES+1]; /* 2019-05 nzx[iy]=nzofs[i]-nzofs[i-1] */
  /* TODO block_nzx[AH] per line or mpi ??? replace nzofs? +xy? */ 
  /* array+1 allows to store initial 0 at end without segfault, simpler code */
  double nhn[NUM_AHEAD_LINES];  /* diagonal element H[n,n] [AHEAD] 2013-02 */
  cplx   nhv[NUM_AHEAD_LINES];  /* 2013-02-07 diag2 _nhv sum_i H[n,i]*|v0i> */
  /* line_start: pointer to first matrix element of the actual matrix line */
  /* old struct_shelem.{bj,blkj,jj,rr}.el[NZX*NAH] bad for SIMD 2014     */
  tbase  acj[HBLrVS]; /* nzcfgj=array config |j>, reused for scfg, 1st=diag */
  int    abj[HBLrVS]; /* nzblkj=abj=array_block_of_j=dest_node_or_task, rename to nj? */
  /* 2017-04 replace ULong by XY_TYPE, ToDo18: put b_hbuf(.xy) here */
  XY_TYPE jj[HBLrVS]; /* nzidxj=jj=scfg2idx=0..b_len[blk]? +XY_FLAG =XY_TYPE? */
  cplx    rr[HBLrVS]; /* nzHij=rr=<iHj>, store disk compressed */
  /* ToDo: char para;  parameter index J1 J2 etc, long name for search/grep */
  tphase asj[HBLrVS]; /* temporary minsymcfg.sign */
  Tnorm2 an2[HBLrVS]; /* temporary minsymcfg.norm2 */
  /* ToDo18: add pointer to (struct thxy) b_hbuf + cpu-local ini touch */
  //    thxy .nz[HBLen] = packed struct .xy .by .rr  in h_file.h
  //    HBLen = ((NZXMAX)*(NUM_AHEAD_LINES))
  struct thxy packed_hxy; /* old18.name=b_hbuf[B_NUM], big size! old=static */
#ifdef CONFIG_MPI /* buffer for remote transfers of nonlocal stored cfgs */
  // ... Isend_remote_cfg + scfg2i_local + Wait + scfg2i_remote ...
  // ToDo17: int remote_dest_cnt[MPI_MAX_NUM];   // num_elements per dest.
  // ToDo17: int remote_dest_ofs[MPI_MAX_NUM+1]; // ofs_elements per dest.
  // ToDo17: tbase remote_cfg[HBLrVS];    // dest_reordered cfgs
#endif
} thbuf; /* struct shbuf, initialize this per thread using alloc_hbuf() */

/*  thbuf *hbuf; hbuf=alloc_hbuf(); ... free_hbuf(hbuf); */
/* ToDo: 2014-07 not needed globally anymore, main() only */
/*
// thbuf *alloc_hbuf( void );
// void  *free_hbuf( thbuf *pb );
*/


/* ================= inline - functions ======================= */

/* this function is only for better readability */
/* function returns up and down spin of site i, 0=empty, 1=up, ... */
static inline int geti(const tbase cfg,const int i){   /* -> 0,1=u,2=d,3=du */
#if CM(1,0,0)
  return 1+(1&(int)(cfg>>i));}     /* 0,1 -> 1,2 */
#else
  return   (3&(int)(cfg>>(2*i)));} /* 0,1,2,3 */
#endif

/* function set up and/or down spin on site i, 0=empty, 1=up, ... */
static inline tbase puti(tbase cfg,const int i,const int s){   /* -> 0,1=u,2=d,3=du */
#if CM(1,0,0)
  return (cfg & ~((tbase)1<<i)) | ((tbase)(s-1))<<i; }
#else
  return (cfg & ~(((tbase)3)<<(2*i))) | ((tbase)s)<<(2*i);}
#endif
        

/* fast encode ud[] to cfg, 0 1=u 2=d 3=du ud[0]=lowest Dez01 */
static inline void encode_ud(tbase *cfg, t_sites ud[NN], int nn){
  int i; *cfg=TBASE0;
#if CM(1,0,0)
  for(i=nn-1;i>=0;i--) (*cfg) = ((*cfg)<<1) | ((tbase)(ud[i]-1));
#else
  for(i=0;i<nn;i++)   (*cfg) |= ((tbase)ud[i])<<(2*i);
#endif
}

#if CM(0,1,1)
/* fast xor single spin i ud=0: u, ud=1: d, Apr02 */
static inline tbase tbase_xor(tbase cfg, int i, int ud){
#if CM(1,0,0)
  cfg ^= (((tbase)1)<<i);    /* ud not used!? */
#else
  cfg ^= (((tbase)1)<<(2*i+ud));
#endif
  return cfg;
}
#endif

int b_out(struct fullsbase *b,int len);
/* 2018-11 renamed paini to binomial_ini */
/* binomial_ini = initialize binomial coefficients, old paini */ 
void   binomial_ini(const int nn);
ULong  numsymconf();
/* const ULong numconf(const int nn, const int nu, const int nd); */
void speed_tests(void *ptr);
void ini_norm2(void);
tbase  ll1(ULong ii,int nn, int nu, int nd); /* only for ll1(0) ??? */
int initmasks(int nn);

int blk_scfg2idx(tbase *cfg, XY_TYPE *idx, int n, thbuf *hbuf); /* more speed!? */
XY_TYPE scfg2idx(tbase cfg);  /* get idx 0...node_n1-1 of symcfg */
int   scfg2node(tbase cfg); /* get taskid 0..mpi_n-1 of symcfg */
#define SZccfg  ((S_tbase)*2+3) /* SIZE = 2 hex digits per byte + pre/suffix */
char *ccfg(const tbase ic); /* Spinconfiguration |..43210> zB: |...o.xoQx> */
char *ccfg2(const tbase ic, char *s1); /* thread safe */

static inline  /* speedup for expectation + tU-sign 2016-01 */
int zi(tbase cfg,int i){   /* Siz -> -1,0,1, moved from spins.c 2016-01 */
#if CM(1,0,0)
  return 1-2*(1&(int)(cfg>>i));}
#else
  int i1=geti(cfg,i); return( (i1==2) ? -1 : (i1==1) ? 1 : 0);}
#endif

static inline
int zizj(tbase cfg,int i,int j){   /* Siz*Sjz -> -1,0,1 */
#if CM(1,0,0)
  return 1-2*(1&(((int)(cfg>>i))^(int)(cfg>>j)));}
#else
  int i1=geti(cfg,i)*geti(cfg,j); return((i1==2)?-1:(i1==4 || i1==1)?1:0);}
#endif


/* 2015-02:  b_getbase renamed to allsymcfg */
tbase allsymcfg(const tbase ib1,struct fullsbase *b);

#if CM(0,1,1)
/*
 * this comes from fermionic sign (-1)^Sum(n(i<j)) for site j
 *   we calculate a product for 2 sites here
 * Is parity base coding dependend? => only mask depends from base code!
 * +1 if even number of 1-bits,  -1 else
 * inline -> code inserted (faster,only -O and C++)
 * static -> function not externaly available    
 */
static inline int parity(tbase1 ib1){ /* odd:-1=1111, even:+1=0001 */
#if NN>CM(64,32,32)
    int i;              /* this works only with class vlint */
    for(i=64;i<CM(nn,2*nn,2*nn);i<<=1) ib1^=ib1>>i; /* not testet ??? */
#endif
#if NN>CM(32,16,16)
    ib1^=ib1>>32;
#endif
    ib1^=ib1>>16; ib1^=ib1>>8; ib1^=ib1>>4;
    ib1^=ib1>>2;  ib1^=ib1<<1; return 1-(2&(int)ib1); /* 6 OPs */
/*  return 1-(((0x6996 >> int(ib1&0xf))<<1) & 2); */  /* 5 OPs */
/*  return 1-( (0xd32c >> int(ib1&0xf))     & 2); */  /* 4 OPs 32bit ToDo */
}
#endif /* CM */

/* see spins.c, depends from base coding (code2)
 *   signum between two sites, for one site use p_mask[j][0] ???
 */
// extern tbase1 p_mask[NN][NN];  /* obsolete ? */
tbase1 get_parity_mask(int ia, int ib); /* replace p_mask ? */

/* return a unique(?) list of counted different pairs = ising energy index
   uu+dd ud+du ..+33 .3+3. .u+.d+u.+d.+3u+3d+u3+d3  (16pairs -> 5sympairs)
   the list is allocated and must be freed after use
   mode==0: output list, free list, return 0
   mode==1: return list
   buf = return space OR NULL for malloc
 */
int *get_dcorr(tbase b1, int mode, int *buf); /* Apr06, v2.25 */

/* --- ToDo: define fastest variant of minsymcfg_* as default here: --- */
/*  this is very CPU and compiler dependent (SIMD, pipelines etc.)
 CPUs have specifique pilelines p.e. 1FusedFltMultAdd 2bMul/Add/BitOp 1SHFT ...
 MultiCoreConnects+MemChannels+Cache limit BW+IPC (Failure-robustness-ECC?)
best-choice table, time in CPU-clks per symmetry, smaller is better 2015-06
CPU \ ALGO        lNv1  lNv8   bn1     bn8 (clks/permutation)
  1ALU^SHFT        160   160   110     81.1 clks 11*(7+3/8)   IPC=1
  1ALU+SHFT         82    80    77     46.8 clks              IPC=1.7
  1ALU^SHFT+1ALU    82    80    55     40.6 clks              IPC=2
  2(ALU^SHFT)       82    80    55     40.6 clks = 1*SSE2-ALU IPC=2
  2ALUs+SHIFT       82          57     39?                    IPC=2.1
  3ALUs             43    54?   40     27.0 clks              IPC=3
  3ALU128           22          20     13.5 clks              IPC=6
  3ALU256           11          10      6.8 clks              IPC=12 (i64)
  1SHFT256+2LOG256+ 21    16?   11      8.8 clks +1shft^alu   IPC=9.2 (i64)

2015-10-08 + factor 2 correction (ToDo: add lNvs? sun?)
# units is CPU-clocks (VECTYPE=0=float32 gives best results, ToDo19, why?)
CPU \ ALGO   lNvs lNb14 19 VS bnVS     bestChoice(benchmarks) IPC=81i/bnVclks
  MIPS-r4k   192   43  47  v4  93   clks lNbrk SiCortex 5Kf IPC=0.9
  AtomN455   141   44  51 v16  44   clks bn16  SSE*8pipe    IPC=1.8 in-order
    pCore=HT 128   32  35  v8  30   clks bn8   2 threads, per core, IPC=2.7
  k8          87   24  30  v8  41   clks lNbrk comp1-2.6GHz=k10exe  IPC=1.9
  k10              24  28  v8  26   clks bn8   comp2-2.3GHz,meggie  IPC=3.1
  epyc  ???                v8            bn8   hpc18-2.4GHz         IPC= ???
  iCore2      60   24  30  v8  23.5 clks bn8   JS-PC-2.7GHz_@1.6GHz IPC=3.4
  sandybrdg.       20  26  v8  20   clks bn8   E54620-2.2GHz        IPC=4.05
  haswell     36   16  20 v32  10.5 clks bn32  AVX2*8pipe           IPC=7.7=max2pipes*4simd*64bit v3=16FLOP64/clk
  # ToDo19 check haswell gcc821
  AVX512+icc       62  70 v16  54   clks bn64? 1thread! XeonPhi IPC=3? bad?
     pC=4HTs  ?                 ?            57*4HT XeonPhi in-order ToDo asm?
  # haswell-advantage=2.4x using vectorized benes-algo, bnVS/lNvs=ca3.5
  # speedtest used, the SH-loop is usually 2-3fold slower (why?)
  # lNbrk: lNb14=14+26 lNb19=19+21
  # itanium 4DP/clk (usually linpack means DP O(n^3), but SP version exist)
  # 3 linpacks: n=100 n=1000 +highly-parallel(HPL)=top500=64bit
 */
#if 1 && NN<=CM(512,0,0) && !defined(__R4000__) // 2015-07-06 NN<=64, ToDo: 512
// && !defined(__k8__)  // k8 is set on other platforms too p.e. sandybridge 
// ToDo: bug in NN>64 (NN=92 ud=6+86 lc.gpl)
#define CFG_USE_BENES 1  /* 0=old_lNbrk_code 1=new_bn_code vectorizable */
#else
#define CFG_USE_BENES 0  /* ToDo: fermionic sign for benes network */
#endif
/* ToDo2015: add tJ, tU, + JJ NN>64, 2015-09-23 test.JJ128 */
#if     CFG_USE_BENES  /* faster for old CPUs and small NN */
#define minsymcfg_dflt           minsymcfg_lNbrk  /* no bn available, unused */
#define minsymcfg_blk_dflt       minsymcfg_blk_bnVS
#define minsymcfg_blk_dflt_no_ud minsymcfg_blk_bnVS_no_ud
#else /* faster on vector CPUs or AVX2 (untested) and big NN (64++) */
#define minsymcfg_dflt           minsymcfg_lNbrk     /* may be slow */
#define minsymcfg_blk_dflt       minsymcfg_blk_lNbrk
#define minsymcfg_blk_dflt_no_ud minsymcfg_blk_lNbrk /* missing _no_ud */
#endif

/* icc16 warns "const" as meaningless, we do not really need it */ 
/* const */
tbase minsymcfg_dflt(const tbase ib1, thbuf *hbuf); /* single config */
void  minsymcfg_blk_dflt(thbuf *hbuf);        /* cfg block (vectorized) */
void  minsymcfg_blk_dflt_no_ud(thbuf *hbuf);
    

#endif
