#!/bin/bash
#
# get M(H) from E(Z) files, version 0.71 May2003 J.Schulenburg
#
#  options: mode [data-file]
#      mode: 0=print only data M(H)
#            1=gnuplot-file (output x11)
#            2=gnuplot-file (output eps to stdout)
#            3=gnuplot-file (output color-eps to stdout)
#            4=gnuplot-file (output dumb)
#          +16=starting gnuplot (output is eps)
#      data-file: list of S and Energy(S) (if omitted use stdin) 
#  example:
#      plateau.sh 17 dat/til8_{18,24,36}.dat  # using bash expansion
#      # til8_{18,24,36}.dat are 3 files for different N
#      #  with 2 columns (1st = Sz = { 0,1,...,N/2 } and 2nd = E0(Sz))
# 
# ToDo: option for get gnuplot-file or pipe to gnuplot
#     wirkt sich $LANG auf awk -lt,-gt aus?
#
# E(S)-hS =< E(S')-hS'
# 
# h_c=min(s',(E(s)-E(s'))/(s-s'))   for all s>s'
#
#  step 1: find min E(s) for h=0; set mag=s(Emin)
#  step 2: find next h for low e(s)
#
# echo "# options: $0 $1 $2 $3"
# avoid german output of 1,03 instead of 1.03
export LANG=C
if test -z "$1"; then echo " error! no mode-option set"; head -15 $0; exit; fi
if test -r $1; then echo " error! no mode-option set"; head -15 $0; exit; fi
declare -x mout=$(( $1 & 16 ))
declare -x mode=$(( $1 & 15 ))
if test -z "$2"; then declare -x file="-"; else
  declare -x file=$2
  if ! test -r $2; then echo " file $2 not found"; exit; fi
fi
if test ! -z "$3"; then
 (
   printf "# gnuplot:\nset xlabel \"h\"\nse ylabel \"m(h)\"\n"
   printf "set key right bottom\n"
   if test $mode = 2; then printf "set term post eps enhanced mono \"Helvetica\" 20\n"; fi
   if test $mode = 3; then printf "set term post eps color solid \"Helvetica\" 20\nset size 1.0,1.0\n"; fi
   if test $mode = 4; then printf "set term dumb\n"; fi
   printf "plot [0:] [0:1] -1 t \"\""
   ( while test ! -z "$2"; do
      shift; x=$(basename $1 .dat)
      printf ",\\\\\n\"-\" t \"$x\" w steps"
     done 
   )
   printf "\n"
   while test ! -z "$2"; do
     sh $0 0 $2
     shift
     printf "e\n"
   done
   if test "1" = "$mode"; then printf "pause 60\n"; fi
 ) | if test "$mout" -gt "0"; then gnuplot; else cat; fi 
 exit
fi
awk -v mode=$mode ' 
  BEGIN { smax=0; smin=0; emin=9999; }
  /^#===no_s===/ { smax = $2/2; printf "# smax=%d\n",smax; }
  (!/^#/) && NF>1 {
    if ( int($1)>smax ) { smax=$1; }
    if ( $2<ee[int($1)] || length(ee[int($1)])==0 )  ee[int($1)]=$2;
    ss[int($1)]=$1;
    if ( $2<emin ) { emin=$2; smin=$1; } 
  }
  END {
   hmin=0
   for (i=0;i<=smax;){
     for (;i<=smin;i++){ hcrit[i]=hmin }  # jump m(h)
     # search for next m(h)-jump == min crit_h(s>smin)
     # printf "#DBG1 %d %d %d %d\n",i,ss[i],smin,smax
     if ( ss[i]-smin<=0 ) { i++; continue; } # 0.3+
     hmin=(ee[i]-emin)/(ss[i]-smin); jmin=i
     for( j=i+1;j<=smax;j++ ){
       if ( ss[j]-smin<=0 ) continue;
       h=(ee[j]-emin)/(ss[j]-smin)
       if (h<=hmin) { hmin=h; jmin=j; }
     }
     # set new minimum for next h-step
     smin=ss[jmin]; emin=ee[jmin]
   }
   # output the result
   print "# lowest energies E(S) found (smin=",smin,",smax=",smax,") :"
   el=ee[0];il=-1;
   for (i=0;i<=smax;i++){
     if( ( i>0 && ss[i]<=0 ) || length(ee[i])==0 ) continue;
     printf "# %5.1f %12.8f %12.8f %12.8f %12.8f\n", ss[i], ee[i], ee[i]-ee[i-1], (ee[i]-el)/(i-il), hcrit[i];
     el=ee[i]; il=i;
   }
   print "# magnetisation curve Z(h)/Zmax=m(h)"
   if ( mode>0 ) {
     printf "# gnuplot:\nset xlabel \"h\"\nse ylabel \"m(h)\"\n"
     if ( mode==2 ) print "set term post eps enhanced mono \"Helvetica\" 20\n"
     if ( mode==3 ) print "set term post eps enhanced color \"Helvetica\" 20\n"
     if ( mode==4 ) print "set term dumb"
     printf "plot [0:] [0:1] \"-\" t \"\" w steps lt 1 lw 2\n"
   }
   for (i=0;i<=smax;i++){
     if( (i>0 && ss[i]<=0) || length(ee[i])==0 || (i>1 && ss[i-1]<=0)){ printf "\n"; continue; }
     # printf "# i=%d ss=%5.1f ee=%12.8f hc=%12.8f\n", i, ss[i], ee[i], hcrit[i];
     printf " %12.8f %8.4f\n",      hcrit[i  ], ss[i]/smax
   } printf " %12.8f %8.4f\n", 1.02*hcrit[i-1], 1
   if ( mode>0  ) printf "e\n"
   if ( mode==1 ) printf "pause 600\n"
  }
' $file | if test "$mout" -gt "0"; then gnuplot; else cat; fi 
